//! Wrapper types to make working with JWTs easier

use std::{array::TryFromSliceError, collections::BTreeMap, fmt::Display, ops::Deref};

use hmac::{Hmac, NewMac};
use jwt::{SignWithKey, VerifyWithKey};
#[cfg(feature = "from_param")]
use rocket::FromForm;
use sha2::Sha256;

/// Wrapper type around a JWT secret to enforce secret length.
#[derive(Debug, Clone, Copy)]
pub struct Secret(pub [u8; 32]);

impl Secret {
    pub fn new(value: &[u8; 32]) -> Self {
        Self(value.to_owned())
    }

    pub fn to_key(&self) -> Hmac<Sha256> {
        // Safety: Unwrap is fine since the key length is always 32.
        Hmac::new_from_slice(&self.0).unwrap()
    }
}

impl Deref for Secret {
    type Target = [u8; 32];

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl TryFrom<&[u8]> for Secret {
    type Error = TryFromSliceError;

    fn try_from(value: &[u8]) -> Result<Self, Self::Error> {
        Ok(Self(value.try_into()?))
    }
}

/// Struct for easier handling of json web tokens
#[derive(Debug, Clone, Default)]
pub struct Token {
    claims: BTreeMap<String, String>,
}

impl Token {
    pub fn new() -> Self {
        Self::default()
    }

    /// Parse a token string into a [`Token`]
    pub fn from_token_string(token: &str, secret: &Secret) -> Result<Self, jwt::error::Error> {
        let claims: BTreeMap<String, String> = token.verify_with_key(&secret.to_key())?;

        Ok(Self { claims })
    }

    /// Consume `self` and return the signed token string.
    pub fn sign(self, secret: &Secret) -> Result<SignedToken, jwt::error::Error> {
        match self.claims.sign_with_key(&secret.to_key()) {
            Ok(token) => Ok(SignedToken(token)),
            Err(e) => Err(e),
        }
    }

    /// Add a claim to the Token
    pub fn add_claim(&mut self, key: impl Into<String>, value: impl Into<String>) {
        self.claims.insert(key.into(), value.into());
    }

    /// Remove a claim from the Token, returning a value if it was present.
    pub fn remove_claim(&mut self, key: &str) -> Option<String> {
        self.claims.remove(key)
    }

    /// Get a ref to a claim in the Token
    pub fn claim(&self, key: &str) -> Option<&String> {
        self.claims.get(key)
    }

    /// Remove all claims from the Token
    pub fn clear_claims(&mut self) {
        self.claims.clear()
    }
}

/// Wrapper type for a signed JWT.
#[derive(Debug, Clone)]
pub struct SignedToken(String);

impl SignedToken {
    pub fn as_str(&self) -> &str {
        &self.0
    }
}

impl Display for SignedToken {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl Deref for SignedToken {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

/// Wrapper type for an unverified token string.
///
/// If the `from_param` crate feature is enabled then this will derive
/// [`rocket::request::FromParam`] and [`rocket::FromForm`]
#[derive(Debug, Clone)]
#[cfg_attr(feature = "from_param", derive(FromForm))]
pub struct UnverifiedToken(String);

impl UnverifiedToken {
    pub fn new(value: String) -> Self {
        Self(value)
    }

    /// Verify an unverified token into a [`Token`]
    pub fn verify(&self, secret: &Secret) -> Result<Token, jwt::error::Error> {
        Token::from_token_string(&self.0, secret)
    }

    pub fn as_str(&self) -> &str {
        &self.0
    }
}

#[cfg(feature = "from_param")]
mod from_param {
    use super::*;

    use rocket::request::FromParam;

    impl FromParam<'_> for UnverifiedToken {
        type Error = ();

        fn from_param(param: &str) -> Result<Self, Self::Error> {
            Ok(UnverifiedToken(param.to_owned()))
        }
    }
}
